@@ -1,6 +1,6 @@
-#![allow(clippy::transmute_ptr_to_ptr, clippy::zero_ptr)] // suppress warnings in py_class invocation
+#![allow(clippy::transmute_ptr_to_ptr, clippy::zero_ptr, clippy::manual_strip)] // suppress warnings in py_class invocation
 use cpython::{
-    NoArgs, ObjectProtocol, PyClone, PyObject, PyResult, PyTuple, Python, PythonObject, ToPyObject,
+    NoArgs, ObjectProtocol, PyObject, PyResult, PyTuple, Python, PythonObject, ToPyObject,
 };
 use errno::errno;
 use log::{debug, error};
@@ -9,24 +9,25 @@ use std::cmp;
 use std::io::Error;
 use std::os::unix::io::{AsRawFd, RawFd};
 
-use crate::pyutils::{close_pyobject, with_released_gil, PyGILState_STATE};
+use crate::pyutils::{close_pyobject, with_released_gil};
 use crate::transport::would_block;
 
 // This is the maximum the Linux kernel will write in a single sendfile call.
-const SENDFILE_MAXSIZE: usize = 0x7fff_f000;
+#[cfg(target_os = "linux")]
+const SENDFILE_MAXSIZE: isize = 0x7fff_f000;
 
 pub struct SendFileInfo {
-    pub content_length: Option<usize>,
-    pub blocksize: i64,
+    pub content_length: isize,
+    pub blocksize: isize,
     pub offset: libc::off_t,
     pub fd: RawFd,
     pub done: bool,
 }
 
 impl SendFileInfo {
-    pub fn new(fd: RawFd, blocksize: i64) -> Self {
+    pub fn new(fd: RawFd, blocksize: isize) -> Self {
         Self {
-            content_length: None,
+            content_length: -1,
             blocksize,
             offset: 0,
             fd,
@@ -35,18 +36,21 @@ impl SendFileInfo {
     }
 
     // true: chunk written completely, false: there's more
-    pub fn send_file(&mut self, out: &mut dyn AsRawFd) -> bool {
+    #[cfg(target_os = "linux")]
+    pub fn send_file(&mut self, out: &mut dyn AsRawFd) -> (bool, usize) {
         debug!("Sending file");
         let mut count = if self.blocksize < 0 {
             SENDFILE_MAXSIZE
         } else {
-            self.blocksize as usize
+            self.blocksize
         };
-        if let Some(cl) = self.content_length {
-            count = cmp::min(cl - self.offset as usize, count);
+        if self.content_length >= 0 {
+            count = cmp::min(self.content_length - self.offset as isize, count);
         }
         self.done = (count == 0) || {
-            match unsafe { libc::sendfile(out.as_raw_fd(), self.fd, &mut self.offset, count) } {
+            match unsafe {
+                libc::sendfile(out.as_raw_fd(), self.fd, &mut self.offset, count as usize)
+            } {
                 -1 => {
                     // will cover the case where count is too large as EOVERFLOW
                     // s. sendfile(2)
@@ -60,18 +64,61 @@ impl SendFileInfo {
                 }
                 // 0 bytes written, assuming we're done
                 0 => true,
+                _ if (self.content_length > 0) => self.content_length == self.offset as isize,
                 // If no content length is given, num_written might be less than count.
                 // However the subsequent call will write 0 bytes -> done.
-                num_written => num_written == count as isize,
+                _ => false,
             }
         };
-        self.done
+        (self.done, self.offset as usize)
     }
 
-    fn update_content_length(&mut self, content_length: usize) {
-        self.content_length = Some(content_length);
-        if self.blocksize > content_length as i64 {
-            self.blocksize = content_length as i64;
+    #[cfg(target_os = "macos")]
+    pub fn send_file(&mut self, out: &mut dyn AsRawFd) -> (bool, usize) {
+        debug!("Sending file");
+        let mut count: i64 = cmp::max(0, self.blocksize as i64);
+        if (self.content_length > 0) && (count > 0) {
+            count = cmp::min(self.content_length as i64 - self.offset, count);
+        }
+        self.done = {
+            let res = unsafe {
+                libc::sendfile(
+                    self.fd,
+                    out.as_raw_fd(),
+                    self.offset,
+                    &mut count,
+                    std::ptr::null_mut(),
+                    0,
+                )
+            };
+            if count == 0 {
+                true
+            } else {
+                self.offset += count;
+                if res == -1 {
+                    let err = Error::from(errno());
+                    if !would_block(&err) {
+                        error!("Could not sendfile(): {:?}", err);
+                        true
+                    } else {
+                        false
+                    }
+                } else {
+                    if self.content_length > 0 {
+                        self.content_length <= self.offset as isize
+                    } else {
+                        false
+                    }
+                }
+            }
+        };
+        (self.done, self.offset as usize)
+    }
+
+    fn update_content_length(&mut self, content_length: isize) {
+        self.content_length = content_length;
+        if self.blocksize > content_length {
+            self.blocksize = content_length;
         }
     }
 }
@@ -92,12 +139,9 @@ py_class!(pub class FileWrapper |py| {
     data filelike: RefCell<PyObject>;
     data sendfileinfo: RefCell<SendFileInfo>;
 
-    def __new__(_cls, filelike: PyObject, blocksize: Option<i64>=None) -> PyResult<Self> {
+    def __new__(_cls, filelike: PyObject, blocksize: Option<isize>=None) -> PyResult<Self> {
         let mut filelike = RefCell::new(filelike);
-        let blocksize = match blocksize {
-            Some(bs) => bs,
-            None => -1
-        };
+        let blocksize = blocksize.unwrap_or(-1);
         let mut fd: RawFd = -1;
         if let Ok(fdpyob) = filelike.get_mut().call_method(py, "fileno", NoArgs, None) {
             if let Ok(pyfd) = fdpyob.extract(py) {
@@ -120,10 +164,11 @@ py_class!(pub class FileWrapper |py| {
 pub trait SendFile {
     // Put this in a trait for more flexibility.
     fn sendfileinfo<'p>(&'p self, py: Python<'p>) -> &'p RefCell<SendFileInfo>;
-    fn send_file(&mut self, out: &mut dyn AsRawFd, gilstate: PyGILState_STATE, py: Python) -> bool;
+    fn send_file(&mut self, out: &mut dyn AsRawFd, py: Python) -> (bool, usize);
     fn update_content_length(&mut self, content_length: usize, py: Python);
     // XXX used only for testing
-    fn new(py: Python, fd: RawFd, bs: i64) -> PyResult<FileWrapper>;
+    #[allow(clippy::new_ret_no_self)]
+    fn new(py: Python, fd: RawFd, bs: isize) -> PyResult<FileWrapper>;
 }
 
 impl SendFile for FileWrapper {
@@ -132,19 +177,17 @@ impl SendFile for FileWrapper {
         self.sendfileinfo(py)
     }
 
-    fn send_file(&mut self, out: &mut dyn AsRawFd, gilstate: PyGILState_STATE, py: Python) -> bool {
-        with_released_gil(gilstate, || {
-            self.sendfileinfo(py).borrow_mut().send_file(out)
-        })
+    fn send_file(&mut self, out: &mut dyn AsRawFd, py: Python) -> (bool, usize) {
+        with_released_gil(|_threadstate| self.sendfileinfo(py).borrow_mut().send_file(out))
     }
 
     fn update_content_length(&mut self, content_length: usize, py: Python) {
         self.sendfileinfo(py)
             .borrow_mut()
-            .update_content_length(content_length);
+            .update_content_length(content_length as isize);
     }
 
-    fn new(py: Python, fd: RawFd, bs: i64) -> PyResult<FileWrapper> {
+    fn new(py: Python, fd: RawFd, bs: isize) -> PyResult<FileWrapper> {
         FileWrapper::create_instance(
             py,
             RefCell::new(py.None()),
@@ -171,8 +214,7 @@ impl Iterator for FileWrapper {
                     .borrow()
                     .blocksize
                     .to_py_object(py)
-                    .as_object()
-                    .clone_ref(py)],
+                    .into_object()],
             ),
             None,
         ) {
@@ -212,8 +254,6 @@ mod tests {
     use cpython::{
         NoArgs, ObjectProtocol, PyClone, PyDict, PyErr, PyTuple, Python, PythonObject, ToPyObject,
     };
-    use env_logger;
-    use log::debug;
     use std::io::{Read, Seek, Write};
     use std::net::{SocketAddr, TcpListener, TcpStream};
     use std::os::unix::io::{AsRawFd, RawFd};
@@ -223,10 +263,6 @@ mod tests {
 
     use crate::filewrapper::{FileWrapper, SendFile};
 
-    fn init() {
-        let _ = env_logger::builder().is_test(true).try_init();
-    }
-
     #[test]
     fn test_no_fileno() {
         let gil = Python::acquire_gil();
@@ -279,7 +315,6 @@ f = FL()"#,
                     for chunk in vec![b"Fo", b"o ", b"42"] {
                         match fw.next() {
                             Some(got) => {
-                                debug!("Got: {:?}", got);
                                 assert_eq!(chunk, &got[..]);
                             }
                             None => {
@@ -419,7 +454,6 @@ f = FL()"#,
 
     #[test]
     fn test_send_file() {
-        init();
         let gil = Python::acquire_gil();
         let py = gil.python();
         let addr: SocketAddr = "127.0.0.1:0".parse().expect("Failed to parse address");
@@ -451,31 +485,27 @@ f = FL()"#,
         });
         let mut connection = TcpStream::connect(addr).expect("Failed to connect");
         let mut sfi = fw.sendfileinfo(py).borrow_mut();
+        let empty: Option<Vec<u8>> = Some(Vec::new());
         sfi.send_file(&mut connection);
         let mut b = got.recv().unwrap();
-        debug!("send_file chunk 1: {:?}", b);
         assert_eq!(&b[..], b"Hell\0\0\0\0\0\0\0\0\0");
         assert_eq!(sfi.offset, 4);
-        assert_eq!(sfi.next(), None);
+        assert_eq!(sfi.next(), empty);
         sfi.send_file(&mut connection);
         b = got.recv().unwrap();
-        debug!("send_file chunk 2: {:?}", b);
         assert_eq!(&b[..], b"o Wo\0\0\0\0\0\0\0\0\0");
         assert_eq!(sfi.offset, 8);
-        assert_eq!(sfi.next(), None);
+        assert_eq!(sfi.next(), empty);
         sfi.send_file(&mut connection);
         b = got.recv().unwrap();
-        debug!("send_file chunk 3: {:?}", b);
         assert_eq!(&b[..], b"rld!\0\0\0\0\0\0\0\0\0");
         assert_eq!(sfi.offset, 12);
-        assert_eq!(sfi.next(), None);
+        assert_eq!(sfi.next(), empty);
         sfi.send_file(&mut connection);
         b = got.recv().unwrap();
-        debug!("send_file chunk 4: {:?}", b);
         assert_eq!(&b[..], b"\n\0\0\0\0\0\0\0\0\0\0\0\0");
         assert_eq!(sfi.offset, 13);
         // no content length + blocksize > number bytes written, next should yield some
-        let empty: Option<Vec<u8>> = Some(Vec::new());
         assert_eq!(sfi.next(), empty);
         sfi.send_file(&mut connection);
         assert_eq!(sfi.next(), None);
@@ -485,7 +515,6 @@ f = FL()"#,
 
     #[test]
     fn test_send_file_updated_content_length() {
-        init();
         let gil = Python::acquire_gil();
         let py = gil.python();
         let addr: SocketAddr = "127.0.0.1:0".parse().expect("Failed to parse address");
@@ -514,13 +543,12 @@ f = FL()"#,
         let mut sfi = fw.sendfileinfo(py).borrow_mut();
         sfi.send_file(&mut connection);
         let mut b = got.recv().unwrap();
-        debug!("send_file chunk 1: {:?}", b);
+        let empty: Option<Vec<u8>> = Some(Vec::new());
         assert_eq!(&b[..], b"Hell\0\0\0\0\0\0\0\0\0");
         assert_eq!(sfi.offset, 4);
-        assert_eq!(sfi.next(), None);
+        assert_eq!(sfi.next(), empty);
         sfi.send_file(&mut connection);
         b = got.recv().unwrap();
-        debug!("send_file chunk 2: {:?}", b);
         assert_eq!(&b[..], b"o\0\0\0\0\0\0\0\0\0\0\0\0");
         assert_eq!(sfi.offset, 5);
         assert_eq!(sfi.next(), None);
@@ -532,7 +560,6 @@ f = FL()"#,
 
     #[test]
     fn test_send_file_content_length_lt_blocksize() {
-        init();
         let gil = Python::acquire_gil();
         let py = gil.python();
         let addr: SocketAddr = "127.0.0.1:0".parse().expect("Failed to parse address");
@@ -559,7 +586,6 @@ f = FL()"#,
         let mut connection = TcpStream::connect(addr).expect("Failed to connect");
         sfi.send_file(&mut connection);
         let b = got.recv().unwrap();
-        debug!("send_file chunk 1: {:?}", b);
         assert_eq!(&b[..], b"Hello\0\0\0\0\0\0\0\0");
         assert_eq!(sfi.offset, 5);
         assert_eq!(sfi.next(), None);
@@ -571,7 +597,6 @@ f = FL()"#,
 
     #[test]
     fn test_file_wrapper_new_no_args() {
-        init();
         let gil = Python::acquire_gil();
         let py = gil.python();
         let fwtype = py.get_type::<FileWrapper>();
