@@ -1,8 +1,6 @@
 #[macro_use]
 extern crate cpython;
 extern crate python3_sys;
-#[macro_use(slog_o)]
-extern crate slog;
 mod filewrapper;
 mod globals;
 mod pymodule;
