@@ -2,52 +2,65 @@ use cpython::{
     NoArgs, ObjectProtocol, PyBytes, PyClone, PyDict, PyResult, PyTuple, Python, PythonObject,
     ToPyObject,
 };
+use encoding::all::ISO_8859_1;
+use encoding::{DecoderTrap, Encoding};
 use httparse::{self};
 use log::{debug, error};
-use std::sync::Arc;
 use urlencoding::decode;
 
-use crate::globals::WSGIGlobals;
-use crate::response::WSGIResponse;
+use crate::globals::SharedWSGIOptions;
 
 // https://tools.ietf.org/html/rfc7230#section-3.2
 // Each header field consists of a case-insensitive field name ...
-pub const CONTENT_LENGTH_HEADER: &str = "content-length";
-const CONTENT_TYPE_HEADER: &str = "content-type";
+pub const CONTENT_LENGTH_HEADER: &str = "CONTENT-LENGTH";
+const CONTENT_TYPE_HEADER: &str = "CONTENT-TYPE";
+const EXPECT_HEADER: &str = "EXPECT";
+const CONTINUE_EXPECTATION: &str = "100-CONTINUE";
 const CONTENT_TYPE: &str = "CONTENT_TYPE";
 const REQUEST_METHOD: &str = "REQUEST_METHOD";
 const PATH_INFO: &str = "PATH_INFO";
 const QUERY_STRING: &str = "QUERY_STRING";
 const SERVER_PROTOCOL: &str = "SERVER_PROTOCOL";
 
+macro_rules! latin1_decode {
+    ($val: expr) => {
+        ISO_8859_1.decode($val, DecoderTrap::Strict)
+    };
+}
+
 #[derive(PartialEq, Debug)]
 pub enum ParsingStage {
     NotParsed,
     HeadersSuccess,
     HeadersError, // parsing headers returned an error, covering the too many headers case
+    Expect100Continue, // the client expects a 100 Continue response before sending the complete request
     ContentPartial,
     ContentComplete,
 }
 
 impl ParsingStage {
     pub fn headers_complete(&self) -> bool {
-        (self != &ParsingStage::NotParsed) && (self != &ParsingStage::HeadersError)
+        (self != &ParsingStage::NotParsed)
+            && (self != &ParsingStage::HeadersError)
+            && (self != &ParsingStage::Expect100Continue)
     }
 
     pub fn complete(&self) -> bool {
         (self == &ParsingStage::ContentComplete) | (self == &ParsingStage::HeadersError)
     }
+
+    pub fn expect_100_continue(&self) -> bool {
+        self == &ParsingStage::Expect100Continue
+    }
 }
 
 pub struct WSGIRequest {
     pub data: Vec<u8>,
-    pub body: Vec<u8>,
     pub stage: ParsingStage,
     pub content_length: usize,
     pub content_start: usize,
     pub http_headers: Vec<(String, String)>,
     pub peer_addr: String,
-    pub response: Arc<Option<WSGIResponse>>,
     num_headers: usize,
 }
 
@@ -55,53 +68,62 @@ impl WSGIRequest {
     pub fn new(num_headers: usize, peer_addr: String) -> WSGIRequest {
         WSGIRequest {
             data: Vec::new(),
-            body: Vec::new(),
             stage: ParsingStage::NotParsed,
             content_length: 0,
             content_start: 0,
             http_headers: Vec::new(),
             peer_addr,
-            response: Arc::new(None),
             num_headers,
         }
     }
 
     #[inline]
     pub fn append(&mut self, data: &[u8]) {
-        self.data.append(&mut data.to_vec());
+        self.data.extend_from_slice(data);
     }
 
     fn parse_headers(&mut self) -> ParsingStage {
         let mut headers = vec![httparse::EMPTY_HEADER; self.num_headers];
         let mut req = httparse::Request::new(&mut headers);
         let length = self.data.len();
+        let mut expect_100_continue = false;
         match req.parse(&self.data) {
             Ok(res) => {
                 if let httparse::Status::Complete(size) = res {
                     self.content_start = size;
                     self.content_length = length - self.content_start;
                     for header in req.headers.iter() {
-                        match header.name.to_lowercase().as_str() {
+                        let uname = header.name.to_ascii_uppercase();
+                        match uname.as_str() {
                             CONTENT_LENGTH_HEADER => {
-                                if let Ok(val) = std::str::from_utf8(header.value) {
+                                if let Ok(val) = latin1_decode!(header.value) {
                                     if let Ok(parsedval) = val.parse() {
                                         self.content_length = parsedval;
                                     }
                                 }
                             }
+                            EXPECT_HEADER => match latin1_decode!(header.value) {
+                                Ok(val) if val.to_ascii_uppercase() == CONTINUE_EXPECTATION => {
+                                    expect_100_continue = true
+                                }
+                                _ => {
+                                    error!(
+                                        "Could not parse Expect header, error value: {:?}",
+                                        header.value
+                                    );
+                                    return ParsingStage::HeadersError;
+                                }
+                            },
                             &_ => {
-                                let key =
-                                    if header.name.to_lowercase().as_str() == CONTENT_TYPE_HEADER {
-                                        CONTENT_TYPE.to_string()
-                                    } else {
-                                        let mut key = "HTTP_".to_string();
-                                        key.push_str(
-                                            &header.name.to_ascii_uppercase().replace("-", "_"),
-                                        );
-                                        key
-                                    };
-                                match String::from_utf8(header.value.to_vec()) {
-                                    Ok(val) => self.http_headers.push((key, val)),
+                                let key = if uname.as_str() == CONTENT_TYPE_HEADER {
+                                    CONTENT_TYPE.to_string()
+                                } else {
+                                    let mut key = "HTTP_".to_string();
+                                    key.push_str(&uname.replace("-", "_"));
+                                    key
+                                };
+                                match latin1_decode!(header.value) {
+                                    Ok(val) => self.http_headers.push((key, val.to_string())),
                                     Err(e) => {
                                         error!("{:?} encountered for value: {:?}", e, header.value);
                                         return ParsingStage::HeadersError;
@@ -115,24 +137,29 @@ impl WSGIRequest {
                             .push((REQUEST_METHOD.to_string(), method.to_string()));
                     }
                     if let Some(path) = req.path {
-                        match decode(path) {
-                            Ok(decoded) => {
-                                let parts: Vec<&str> = decoded.split('?').collect();
-                                self.http_headers
-                                    .push((PATH_INFO.to_string(), parts[0].to_string()));
-                                if parts.len() > 1 {
-                                    self.http_headers
-                                        .push((QUERY_STRING.to_string(), parts[1].to_string()));
-                                } else {
-                                    self.http_headers
-                                        .push((QUERY_STRING.to_string(), "".to_string()));
+                        let parts: Vec<&str> = path.splitn(2, '?').collect();
+                        match decode(parts[0]) {
+                            Ok(path_info) => match latin1_decode!(path_info.as_bytes()) {
+                                Ok(latin1_path_info) => self
+                                    .http_headers
+                                    .push((PATH_INFO.to_string(), latin1_path_info)),
+                                Err(e) => {
+                                    error!("Could not urldecode path info: {:?}", e);
+                                    return ParsingStage::HeadersError;
                                 }
-                            }
+                            },
                             Err(e) => {
                                 error!("Could not urldecode path info: {:?}", e);
                                 return ParsingStage::HeadersError;
                             }
                         }
+                        if parts.len() > 1 {
+                            self.http_headers
+                                .push((QUERY_STRING.to_string(), parts[1].to_string()));
+                        } else {
+                            self.http_headers
+                                .push((QUERY_STRING.to_string(), "".to_string()));
+                        }
                     }
                     if let Some(version) = req.version {
                         if version > 1 {
@@ -144,7 +171,11 @@ impl WSGIRequest {
                                 .push((SERVER_PROTOCOL.to_string(), protocol));
                         }
                     }
-                    ParsingStage::HeadersSuccess
+                    if !self.stage.expect_100_continue() && expect_100_continue {
+                        ParsingStage::Expect100Continue
+                    } else {
+                        ParsingStage::HeadersSuccess
+                    }
                 } else {
                     ParsingStage::NotParsed
                 }
@@ -161,7 +192,7 @@ impl WSGIRequest {
             debug!("Parsing request headers: {:?}", self.stage);
             self.stage = self.parse_headers();
         }
-        if self.stage.headers_complete() {
+        if self.stage.headers_complete() && !self.stage.expect_100_continue() {
             let length = self.data.len();
             self.stage = if self.content_length > length - self.content_start {
                 // expecting more data, maybe from the next read
@@ -171,72 +202,50 @@ impl WSGIRequest {
                 );
                 ParsingStage::ContentPartial
             } else {
-                if self.content_length > 0 {
-                    self.body.extend_from_slice(
-                        &self.data[self.content_start..self.content_start + self.content_length],
-                    );
-                }
                 ParsingStage::ContentComplete
             }
         }
         self.stage.complete()
     }
 
-    pub fn wsgi_environ(&self, globals: &WSGIGlobals, py: Python) -> PyResult<PyDict> {
+    pub fn wsgi_environ(&self, globals: SharedWSGIOptions, py: Python) -> PyResult<PyDict> {
         let io = &globals.io_module;
-        let sys = &globals.sys_module;
-        let environ = PyDict::new(py);
+        let environ = globals.wsgi_environ.copy(py)?;
         for (k, v) in self.http_headers.iter() {
             environ.set_item(py, k, v)?;
         }
-        environ.set_item(py, "SERVER_NAME", globals.server_name)?;
-        environ.set_item(py, "SERVER_PORT", globals.server_port)?;
-        environ.set_item(py, "SCRIPT_NAME", globals.script_name)?;
-        environ.set_item(py, "REMOTE_ADDR", &self.peer_addr)?;
+        environ.set_item(py, globals.peer_addr_key.clone_ref(py), &self.peer_addr)?;
         let input = io.call(py, "BytesIO", NoArgs, None)?;
-        let bodylen = self.body.len();
-        if bodylen > 0 {
-            environ.set_item(py, "CONTENT_LENGTH", bodylen)?;
+        if self.content_length > 0 {
+            environ.set_item(
+                py,
+                globals.content_length_key.clone_ref(py),
+                self.content_length,
+            )?;
             input.call_method(
                 py,
                 "write",
                 PyTuple::new(
                     py,
-                    &[PyBytes::new(py, &self.body[..]).as_object().clone_ref(py)],
+                    &[PyBytes::new(
+                        py,
+                        &self.data[self.content_start..self.content_start + self.content_length],
+                    )
+                    .into_object()],
                 ),
                 None,
             )?;
             input.call_method(
                 py,
                 "seek",
-                PyTuple::new(py, &[(0 as i32).to_py_object(py).as_object().clone_ref(py)]),
+                PyTuple::new(py, &[(0_i32).to_py_object(py).into_object()]),
                 None,
             )?;
         }
-        environ.set_item(py, "wsgi.input", input)?;
-        environ.set_item(py, "wsgi.errors", sys.get(py, "stderr")?)?;
-        environ.set_item(py, "wsgi.version", (1, 0))?;
-        environ.set_item(py, "wsgi.multithread", false)?;
-        environ.set_item(py, "wsgi.multiprocess", true)?;
-        environ.set_item(py, "wsgi.run_once", false)?;
-        environ.set_item(py, "wsgi.url_scheme", "http")?;
-        if let Some(wsgi) = &globals.wsgi_module {
-            debug!("Setting FileWrapper in environ");
-            environ.set_item(py, "wsgi.file_wrapper", wsgi.get(py, "FileWrapper")?)?;
-        }
+        environ.set_item(py, globals.wsgi_input_key.clone_ref(py), input)?;
         Ok(environ)
     }
 
-    #[inline]
-    pub fn get_response(&mut self) -> Option<&mut WSGIResponse> {
-        if let Some(opt) = Arc::get_mut(&mut self.response) {
-            if let Some(resp) = opt {
-                return Some(resp);
-            }
-        }
-        None
-    }
-
     pub fn is_bad_request(&self) -> bool {
         self.stage == ParsingStage::HeadersError
     }
@@ -247,11 +256,9 @@ mod tests {
     use cpython::{NoArgs, ObjectProtocol, PyBytes, PyClone, PyString, Python, PythonObject};
     use log::debug;
     use std::net::SocketAddr;
-    use std::sync::Arc;
 
-    use crate::globals::WSGIGlobals;
+    use crate::globals::{shared_wsgi_options, SharedWSGIOptions};
     use crate::request::{ParsingStage, WSGIRequest};
-    use crate::response::WSGIResponse;
 
     macro_rules! assert_header {
         ($got:ident, $py:ident, $key:literal, $value:expr) => {
@@ -267,12 +274,12 @@ mod tests {
         };
     }
 
-    fn make_globals(py: Python) -> (WSGIGlobals, SocketAddr) {
+    fn make_globals(py: Python) -> (SharedWSGIOptions, SocketAddr) {
         let server_name = "127.0.0.1";
         let port = "0";
-        let sn = "/foo";
+        let sn = String::from("/foo");
         (
-            WSGIGlobals::new(server_name, port, sn, py),
+            shared_wsgi_options(String::from(server_name), String::from(port), sn, false, py),
             (server_name.to_string() + ":" + port).parse().unwrap(),
         )
     }
@@ -309,7 +316,7 @@ mod tests {
 
     #[test]
     fn test_url_decode() {
-        let raw = b"GET /foo%2042?bar=baz%20foo HTTP/1.1\r\nHost: localhost:7878\r\nUser-Agent: Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:70.0) Gecko/20100101 Firefox/70.0\r\nAccept: image/webp,*/*\r\nAccept-Language: de-DE,en-US;q=0.7,en;q=0.3\r\nAccept-Encoding: gzip, deflate\r\nConnection: keep-alive\r\nCookie: foo_language=en;\r\nDNT: 1\r\n\r\n";
+        let raw = b"GET /foo%2042?bar=baz%20foo&next=newsletter%3D%252F434%252F%252F2021-03-05%26rw%3Dtrue HTTP/1.1\r\nHost: localhost:7878\r\nUser-Agent: Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:70.0) Gecko/20100101 Firefox/70.0\r\nAccept: image/webp,*/*\r\nAccept-Language: de-DE,en-US;q=0.7,en;q=0.3\r\nAccept-Encoding: gzip, deflate\r\nConnection: keep-alive\r\nCookie: foo_language=en;\r\nDNT: 1\r\n\r\n";
         let mut got = WSGIRequest::new(16, String::new());
         got.append(raw);
         got.parse_data();
@@ -318,11 +325,47 @@ mod tests {
             match name.as_str() {
                 "HTTP_COOKIE" => assert!(&value[..] == "foo_language=en;"),
                 "PATH_INFO" => assert!(&value[..] == "/foo 42"),
-                "QUERY_STRING" => assert!(&value[..] == "bar=baz foo"),
+                "QUERY_STRING" => assert!(&value[..] == "bar=baz%20foo&next=newsletter%3D%252F434%252F%252F2021-03-05%26rw%3Dtrue"),
+                "HTTP_ACCEPT" => assert!(&value[..] == "image/webp,*/*"),
+                "HTTP_ACCEPT_LANGUAGE" => assert!(&value[..] == "de-DE,en-US;q=0.7,en;q=0.3"),
+                "HTTP_ACCEPT_ENCODING" => assert!(&value[..] == "gzip, deflate"),
+                "HTTP_CONNECTION" => assert!(&value[..] == "keep-alive"),
+                "REQUEST_METHOD" => assert!(&value[..] == "GET"),
+                "HTTP_HOST" => assert!(&value[..] == "localhost:7878"),
+                "HTTP_USER_AGENT" => {
+                    let expected = "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:70.0) Gecko/20100101 Firefox/70.0";
+                    assert_eq!(value, expected);
+                }
+                "HTTP_DNT" => assert_eq!(&value[..], "1"),
+                "SERVER_PROTOCOL" => assert_eq!(&value[..], "HTTP/1.1"),
+                &_ => {}
+            }
+        }
+    }
+
+    #[test]
+    fn test_latin1_url() {
+        let raw = b"GET /f%C3%A4%C3%A4%2042?bar=baz%20foo&next=newsletter%3D%252F434%252F%252F2021-03-05%26rw%3Dtrue HTTP/1.1\r\nHost: localhost:7878\r\nUser-Agent: Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:70.0) Gecko/20100101 Firefox/70.0\r\nAccept: image/webp,*/*\r\nAccept-Language: de-DE,en-US;q=0.7,en;q=0.3\r\nAccept-Encoding: gzip, deflate\r\nConnection: keep-alive\r\nCookie: foo_language=en;\r\nDNT: 1\r\n\r\n";
+        let mut got = WSGIRequest::new(16, String::new());
+        got.append(raw);
+        got.parse_data();
+        assert!(got.http_headers.len() == 12);
+        for (name, value) in got.http_headers.iter() {
+            match name.as_str() {
+                "HTTP_COOKIE" => assert!(&value[..] == "foo_language=en;"),
+                // PATH_INFO contains latin-1 escape characters.
+                // The expected result should be the same as
+                // if we would be using the following Python expression:
+                //
+                // urlib.parse.unquote('/f%C3%A4%C3%A4%2042', encoding='latin-1')
+                //
+                // i.e. the following UTF-8 encoded byte sequence
+                // b'/f\xc3\x83\xc2\xa4\xc3\x83\xc2\xa4 42'
+                "PATH_INFO" => assert!(value.as_bytes() == b"/f\xc3\x83\xc2\xa4\xc3\x83\xc2\xa4 42"),
+                "QUERY_STRING" => assert!(&value[..] == "bar=baz%20foo&next=newsletter%3D%252F434%252F%252F2021-03-05%26rw%3Dtrue"),
                 "HTTP_ACCEPT" => assert!(&value[..] == "image/webp,*/*"),
                 "HTTP_ACCEPT_LANGUAGE" => assert!(&value[..] == "de-DE,en-US;q=0.7,en;q=0.3"),
                 "HTTP_ACCEPT_ENCODING" => assert!(&value[..] == "gzip, deflate"),
-                "HTTP_AUTHORIZATION" => assert!(&value[..] == "Basic YWRtaW46YWRtaW4="),
                 "HTTP_CONNECTION" => assert!(&value[..] == "keep-alive"),
                 "REQUEST_METHOD" => assert!(&value[..] == "GET"),
                 "HTTP_HOST" => assert!(&value[..] == "localhost:7878"),
@@ -366,7 +409,8 @@ mod tests {
                 &_ => {}
             }
         }
-        assert_eq!(&got.body[..], b"field1=value1&field2=value2");
+        let body = &got.data[got.content_start..got.content_start + got.content_length];
+        assert_eq!(body, b"field1=value1&field2=value2");
     }
 
     #[test]
@@ -392,8 +436,42 @@ mod tests {
             }
         }
         let expected = b"field1=value1&field2=value2&field3=value3";
-        debug!("{:?}", got.body);
-        assert!(got.body.iter().zip(expected.iter()).all(|(p, q)| p == q));
+        let body = &got.data[got.content_start..got.content_start + got.content_length];
+        debug!("{:?}", body);
+        assert!(body.iter().zip(expected.iter()).all(|(p, q)| p == q));
+    }
+
+    #[test]
+    fn test_parse_expect_continue() {
+        let raw = b"POST /test HTTP/1.1\r\nHost: foo.example\r\nContent-Type: application/x-www-form-urlencoded\r\nContent-Length: 27\r\nExpect: 100-continue\r\n\r\n";
+        let body = b"field1=value1&field2=value2";
+        let mut got = WSGIRequest::new(16, String::new());
+        got.append(raw);
+        got.parse_data();
+        assert!(!got.stage.headers_complete());
+        assert!(!got.stage.complete());
+        assert!(got.stage.expect_100_continue());
+        got.append(body);
+        got.parse_data();
+        assert!(got.stage.headers_complete());
+        assert!(got.stage.complete());
+        assert!(!got.stage.expect_100_continue());
+        let body = &got.data[got.content_start..got.content_start + got.content_length];
+        assert_eq!(body, b"field1=value1&field2=value2");
+        let raw = b"POST /test HTTP/1.1\r\nHost: foo.example\r\nContent-Type: application/x-www-form-urlencoded\r\nContent-Length: 27\r\nExpect: 101-whatever\r\n\r\n";
+        let mut got = WSGIRequest::new(16, String::new());
+        got.append(raw);
+        got.parse_data();
+        assert!(!got.stage.headers_complete());
+        assert!(got.stage.complete());
+        assert!(!got.stage.expect_100_continue());
+        assert_eq!(got.stage, ParsingStage::HeadersError);
+        got.append(body);
+        got.parse_data();
+        assert!(!got.stage.headers_complete());
+        assert!(got.stage.complete());
+        assert!(!got.stage.expect_100_continue());
+        assert_eq!(got.stage, ParsingStage::HeadersError);
     }
 
     #[test]
@@ -405,7 +483,7 @@ mod tests {
         let mut req = WSGIRequest::new(16, String::from("192.168.1.23"));
         req.append(raw);
         req.parse_data();
-        let got = req.wsgi_environ(&g, py).unwrap();
+        let got = req.wsgi_environ(g, py).unwrap();
         assert_header!(got, py, "SERVER_NAME", "127.0.0.1");
         assert_header!(got, py, "SERVER_PORT", "0");
         assert_header!(got, py, "SCRIPT_NAME", "/foo");
@@ -444,7 +522,7 @@ mod tests {
         let mut req = WSGIRequest::new(16, String::new());
         req.append(raw);
         req.parse_data();
-        let got = req.wsgi_environ(&g, py).unwrap();
+        let got = req.wsgi_environ(g, py).unwrap();
         assert_header!(got, py, "CONTENT_TYPE", "application/x-www-form-urlencoded");
         let input = got
             .get_item(py, "wsgi.input")
@@ -466,7 +544,7 @@ mod tests {
         let mut req = WSGIRequest::new(16, String::new());
         req.append(raw);
         req.parse_data();
-        let got = req.wsgi_environ(&g, py).unwrap();
+        let got = req.wsgi_environ(g, py).unwrap();
         assert_header!(
             got,
             py,
@@ -487,28 +565,13 @@ mod tests {
         assert!(got.iter().zip(expected.iter()).all(|(p, q)| p == q));
     }
 
-    #[test]
-    fn test_get_response() {
-        let mut req = WSGIRequest::new(16, String::new());
-        match req.get_response() {
-            None => (),
-            _ => assert!(false),
-        }
-        let expected = b"Hello";
-        let mut response = WSGIResponse::new();
-        response.current_chunk = expected.to_vec();
-        req.response = Arc::new(Some(response));
-        match req.get_response() {
-            Some(resp) => assert!(&resp.current_chunk == expected),
-            _ => assert!(false),
-        }
-    }
-
     #[test]
     fn test_parsing_stages() {
         assert!(ParsingStage::HeadersSuccess.headers_complete());
         assert!(!ParsingStage::HeadersError.headers_complete());
+        assert!(!ParsingStage::Expect100Continue.headers_complete());
         assert!(!ParsingStage::HeadersSuccess.complete());
         assert!(ParsingStage::ContentComplete.complete());
+        assert!(!ParsingStage::Expect100Continue.complete());
     }
 }
