@@ -1,14 +1,14 @@
 use cpython::{PyObject, Python};
-use log::{debug, error};
+use log::{debug, error, info};
 use mio::{Events, Interest, Poll, Token};
-use python3_sys::{PyEval_RestoreThread, PyEval_SaveThread};
 use std::collections::HashMap;
 use std::io;
 use std::time::Duration;
 
-use crate::globals::WSGIGlobals;
+use crate::globals::ServerOptions;
+use crate::pyutils::with_released_gil;
 use crate::request::WSGIRequest;
-use crate::transport::{would_block, Listener, Read, Result};
+use crate::transport::{would_block, HTTP11Connection, Listener, Listening, Read, Result};
 use crate::workerpool::WorkerPool;
 use crate::workers::{write_blocking_worker, write_non_blocking_worker};
 
@@ -27,35 +27,26 @@ pub struct Server<L: Listener> {
     events: Events,
     listener: L,
     workers: WorkerPool<L::Connected>,
-    connections: HashMap<Token, L::Connected>,
+    connections: HashMap<Token, HTTP11Connection<L::Connected>>,
+    reusable: HashMap<Token, HTTP11Connection<L::Connected>>,
     requests: HashMap<Token, WSGIRequest>,
     current_token: Token,
-    max_number_headers: usize,
+    server_options: ServerOptions,
 }
 
-impl<'g, L: 'static + Listener> Server<L> {
+impl<'g, L: 'static + Listener + Listening> Server<L> {
     pub fn new(
         application: PyObject,
-        globals: &WSGIGlobals<'g>,
         listener: L,
-        num_workers: usize,
-        write_blocking: bool,
-        max_number_headers: usize,
+        server_options: ServerOptions,
         py: Python,
     ) -> io::Result<Server<L>> {
-        let workers = WorkerPool::new(
-            String::from(globals.server_name),
-            String::from(globals.server_port),
-            globals.script_name.to_string(),
-            application,
-            if write_blocking {
-                write_blocking_worker::<L, L::Connected>
-            } else {
-                write_non_blocking_worker::<L, L::Connected>
-            },
-            num_workers,
-            py,
-        );
+        let worker_fn = if server_options.write_blocking {
+            write_blocking_worker::<L, L::Connected>
+        } else {
+            write_non_blocking_worker::<L, L::Connected>
+        };
+        let workers = WorkerPool::new(&server_options, application, worker_fn, py);
         let mut listener = listener;
         let poll = Poll::new()?;
         poll.registry()
@@ -67,9 +58,10 @@ impl<'g, L: 'static + Listener> Server<L> {
             listener,
             workers,
             connections: HashMap::new(),
+            reusable: HashMap::new(),
             requests: HashMap::new(),
             current_token: Token(SERVER.0 + 1),
-            max_number_headers,
+            server_options,
         })
     }
 
@@ -84,7 +76,10 @@ impl<'g, L: 'static + Listener> Server<L> {
             Some(request) => request,
             None => {
                 debug!("creating new WSGI request");
-                WSGIRequest::new(self.max_number_headers, connection.peer_addr())
+                WSGIRequest::new(
+                    self.server_options.max_number_headers,
+                    connection.peer_addr(),
+                )
             }
         };
         // We can (maybe) read from the connection.
@@ -95,7 +90,6 @@ impl<'g, L: 'static + Listener> Server<L> {
                     // Reading 0 bytes means the other side has closed the
                     // connection or is done writing, then so are we.
                     debug!("Reading 0 bytes, consider reading done");
-                    // req.parse_data();
                     connection_closed = true;
                     break;
                 }
@@ -114,7 +108,7 @@ impl<'g, L: 'static + Listener> Server<L> {
             }
         }
 
-        if connection_closed & !req.stage.complete() {
+        if connection_closed & !req.stage.complete() & !req.data.is_empty() {
             debug!("Connection closed");
             Err(io::Error::new(io::ErrorKind::Other, "Incomplete request"))
         } else {
@@ -128,10 +122,16 @@ impl<'g, L: 'static + Listener> Server<L> {
             .poll(&mut self.events, Some(Duration::from_millis(POLL_TIMEOUT)))
         {
             Ok(_) => {
+                for (token, conn) in self.workers.reusable_connections.try_iter() {
+                    self.reusable.insert(token, conn);
+                }
                 for event in self.events.iter() {
                     match event.token() {
                         SERVER => {
-                            while let Ok(mut connection) = self.listener.accept() {
+                            while let Ok(mut connection) = self
+                                .listener
+                                .accept(self.server_options.connection_options.clone())
+                            {
                                 let token = next(&mut self.current_token);
                                 debug!("current token: {:?}", token);
 
@@ -145,7 +145,11 @@ impl<'g, L: 'static + Listener> Server<L> {
                         }
                         token if event.is_readable() => {
                             // (maybe) received an event for a connection.
-                            if let Some(mut connection) = self.connections.remove(&token) {
+                            if let Some(mut connection) = if self.reusable.contains_key(&token) {
+                                self.reusable.remove(&token)
+                            } else {
+                                self.connections.remove(&token)
+                            } {
                                 let req = self.requests.remove(&token);
                                 match self.handle_read_event(&mut connection, req) {
                                     Ok(mut req) => {
@@ -153,6 +157,11 @@ impl<'g, L: 'static + Listener> Server<L> {
                                             debug!("Request complete, relaying to worker pool");
                                             self.workers.execute(token, req, Some(connection))?;
                                         } else {
+                                            if req.stage.expect_100_continue() {
+                                                if let Err(e) = connection.write_100_continue() {
+                                                    error!("Error sending 100 Continue: {:?}", e);
+                                                }
+                                            }
                                             self.requests.insert(token, req);
                                             self.connections.insert(token, connection);
                                         }
@@ -166,6 +175,12 @@ impl<'g, L: 'static + Listener> Server<L> {
                         _ => (),
                     }
                 }
+                // reinsert reusable connection for which there were no events, but increase the
+                // reuse count
+                for (token, conn) in self.reusable.drain() {
+                    debug!("Reusing connection {:?} for token {:?}", conn, token);
+                    self.connections.insert(token, conn);
+                }
             }
             Err(e) => return Err(Box::new(e)),
         }
@@ -173,16 +188,20 @@ impl<'g, L: 'static + Listener> Server<L> {
     }
 
     pub fn serve(&mut self) -> Result<()> {
-        let py_thread_state = unsafe { PyEval_SaveThread() };
-        loop {
-            if let Err(e) = self.poll_once() {
-                error!("Error processing poll events: {:?}", e);
-                self.workers.join()?;
-                break;
+        info!(
+            "pyruvate listening on {}",
+            self.listener.local_addr_string()
+        );
+        with_released_gil(|_threadstate| {
+            loop {
+                if let Err(e) = self.poll_once() {
+                    error!("Error processing poll events: {:?}", e);
+                    self.workers.join()?;
+                    break;
+                }
             }
-        }
-        unsafe { PyEval_RestoreThread(py_thread_state) };
-        Ok(())
+            Ok(())
+        })
     }
 }
 
@@ -190,19 +209,18 @@ impl<'g, L: 'static + Listener> Server<L> {
 mod tests {
 
     use cpython::{PyClone, PyDict, Python, PythonObject};
-    use env_logger;
-    use log::debug;
     use mio::net::TcpListener;
     use mio::Token;
-    use std::io::{self, Write};
+    use std::io::{self, Read as IORead, Write};
     use std::net::{SocketAddr, TcpStream};
     use std::ops::Range;
+    use std::sync::mpsc::channel;
     use std::thread;
     use std::time::Duration;
 
-    use crate::globals::WSGIGlobals;
-    use crate::server::{next, Server};
-    use crate::transport::Read;
+    use crate::globals::{shared_wsgi_options, SharedWSGIOptions};
+    use crate::server::{next, Server, ServerOptions};
+    use crate::transport::{shared_connection_options, Read};
 
     struct StreamMock {
         pub data: Vec<u8>,
@@ -285,18 +303,24 @@ mod tests {
         }
     }
 
-    fn make_globals(py: Python) -> (WSGIGlobals, SocketAddr) {
+    fn make_globals(py: Python) -> (SharedWSGIOptions, SocketAddr) {
         let server_name = "127.0.0.1";
         let port = "0";
-        let sn = "/foo";
+        let sn = String::from("/foo");
         (
-            WSGIGlobals::new(server_name, port, sn, py),
+            shared_wsgi_options(String::from(server_name), String::from(port), sn, false, py),
             (server_name.to_string() + ":" + port).parse().unwrap(),
         )
     }
 
-    fn init() {
-        let _ = env_logger::builder().is_test(true).try_init();
+    fn server_options(globals: SharedWSGIOptions) -> ServerOptions {
+        ServerOptions {
+            num_workers: 2,
+            write_blocking: false,
+            max_number_headers: 16,
+            connection_options: shared_connection_options(10, Duration::from_secs(60)),
+            wsgi_options: globals,
+        }
     }
 
     #[test]
@@ -311,27 +335,24 @@ mod tests {
 
     #[test]
     fn test_handle_read_event() {
-        init();
         let gil = Python::acquire_gil();
         let py = gil.python();
         let (g, si) = make_globals(py);
         let listener = TcpListener::bind(si).unwrap();
-        let server = Server::new(py.None(), &g, listener, 2, false, 16, py).unwrap();
+        let server = Server::new(py.None(), listener, server_options(g), py).unwrap();
         let expected = b"GET /foo42?bar=baz HTTP/1.1\r\nHost: localhost:7878\r\nUser-Agent: Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:70.0) Gecko/20100101 Firefox/70.0\r\nAccept: image/webp,*/*\r\nAccept-Language: de-DE,en-US;q=0.7,en;q=0.3\r\nAccept-Encoding: gzip, deflate\r\nConnection: keep-alive\r\nCookie: foo_language=en;\r\nDNT: 1\r\n\r\n";
         let mut s = StreamMock::new(expected.to_vec(), true, false);
         // read interrupted by WouldBlock, block will occur
         // before complete parsing of request is possible
         match server.handle_read_event(&mut s, None) {
             Ok(mut req) => {
-                debug!("Got (after WouldBlock): {:?}", &req.data[..]);
                 assert!(req.data.len() == 10);
                 assert!(expected.iter().zip(req.data.iter()).all(|(p, q)| p == q));
                 let res = req.parse_data();
                 assert!(req.data.len() == 10);
                 assert!(!res);
             }
-            Err(e) => {
-                debug!("Error: {:?}", e);
+            Err(_) => {
                 assert!(false);
             }
         }
@@ -340,12 +361,10 @@ mod tests {
         // when complete parsing of request is possible
         match server.handle_read_event(&mut s, None) {
             Ok(mut req) => {
-                debug!("Got (after WouldBlock): {:?}", &req.data[..]);
                 assert!(req.data.iter().zip(expected.iter()).all(|(p, q)| p == q));
                 assert!(req.parse_data());
             }
-            Err(e) => {
-                debug!("Error: {:?}", e);
+            Err(_) => {
                 assert!(false);
             }
         }
@@ -366,12 +385,11 @@ mod tests {
 
     #[test]
     fn test_handle_read_event_too_many_headers() {
-        init();
         let gil = Python::acquire_gil();
         let py = gil.python();
         let (g, si) = make_globals(py);
         let listener = TcpListener::bind(si).unwrap();
-        let server = Server::new(py.None(), &g, listener, 2, false, 16, py).unwrap();
+        let server = Server::new(py.None(), listener, server_options(g), py).unwrap();
         let expected = b"GET /foo42?bar=baz HTTP/1.1\r\nHost: localhost:7878\r\nUser-Agent: Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:70.0) Gecko/20100101 Firefox/70.0\r\nAccept: image/webp,*/*\r\nAccept-Language: de-DE,en-US;q=0.7,en;q=0.3\r\nAccept-Encoding: gzip, deflate\r\nConnection: keep-alive\r\nCookie: foo_language=en;\r\nDNT: 1\r\nH1: 1\r\nH2: 2\r\nH3: 3\r\nH4: 4\r\nH5: 5\r\nH6: 6\r\nH7: 7\r\nH8: 8\r\nH9: 9\r\nH10: 10\r\nH11: 11\r\nH12: 12\r\n\r\n";
         let mut s = StreamMock::new(expected.to_vec(), true, false);
         // read interrupted by WouldBlock, block will occur
@@ -380,7 +398,6 @@ mod tests {
             Ok(req) => match server.handle_read_event(&mut s, Some(req)) {
                 Ok(_) => assert!(false),
                 Err(e) => {
-                    debug!("Error: {:?}", e);
                     assert!(e.kind() == io::ErrorKind::Other);
                 }
             },
@@ -390,7 +407,6 @@ mod tests {
 
     #[test]
     fn test_create_blocking_server() {
-        init();
         let gil = Python::acquire_gil();
         let py = gil.python();
         let locals = PyDict::new(py);
@@ -415,12 +431,11 @@ app = simple_app"#,
                     .unwrap()
                     .as_object()
                     .clone_ref(py);
-                match Server::new(app, &g, listener, 2, true, 16, py) {
+                match Server::new(app, listener, server_options(g), py) {
                     Ok(got) => {
                         assert!(got.current_token == Token(1));
                     }
-                    Err(e) => {
-                        debug!("Error when creating Server: {:?}", e);
+                    Err(_) => {
                         assert!(false);
                     }
                 }
@@ -431,7 +446,6 @@ app = simple_app"#,
 
     #[test]
     fn test_create_non_blocking_server() {
-        init();
         let gil = Python::acquire_gil();
         let py = gil.python();
         let locals = PyDict::new(py);
@@ -456,12 +470,11 @@ app = simple_app"#,
                     .unwrap()
                     .as_object()
                     .clone_ref(py);
-                match Server::new(app, &g, listener, 2, false, 16, py) {
+                match Server::new(app, listener, server_options(g), py) {
                     Ok(got) => {
                         assert!(got.current_token == Token(1));
                     }
-                    Err(e) => {
-                        debug!("Error when creating Server: {:?}", e);
+                    Err(_) => {
                         assert!(false);
                     }
                 }
@@ -472,7 +485,6 @@ app = simple_app"#,
 
     #[test]
     fn test_create_non_blocking_server_socket_activation() {
-        init();
         let gil = Python::acquire_gil();
         let py = gil.python();
         let locals = PyDict::new(py);
@@ -497,12 +509,11 @@ app = simple_app"#,
                     .unwrap()
                     .as_object()
                     .clone_ref(py);
-                match Server::new(app, &g, listener, 2, false, 16, py) {
+                match Server::new(app, listener, server_options(g), py) {
                     Ok(got) => {
                         assert!(got.current_token == Token(1));
                     }
-                    Err(e) => {
-                        debug!("Error when creating Server: {:?}", e);
+                    Err(_) => {
                         assert!(false);
                     }
                 }
@@ -513,7 +524,6 @@ app = simple_app"#,
 
     #[test]
     fn test_server_poll_once() {
-        init();
         let gil = Python::acquire_gil();
         let py = gil.python();
         let locals = PyDict::new(py);
@@ -537,16 +547,17 @@ app = simple_app"#,
                     .unwrap()
                     .as_object()
                     .clone_ref(py);
-                let si: SocketAddr = "127.0.0.1:7878".parse().unwrap();
+                let si: SocketAddr = "127.0.0.1:0".parse().unwrap();
                 let listener = TcpListener::bind(si).unwrap();
-                match Server::new(app, &g, listener, 1, true, 16, py) {
+                let addr = listener.local_addr().unwrap();
+                match Server::new(app, listener, server_options(g), py) {
                     Ok(mut got) => {
                         // accept
                         got.poll_once().unwrap();
                         // create request in separate thread
                         let t = thread::spawn(move || {
                             let mut connection =
-                                TcpStream::connect_timeout(&si, Duration::from_millis(1000))
+                                TcpStream::connect_timeout(&addr, Duration::from_millis(1000))
                                     .expect("Failed to connect to server");
                             let req = b"GET /foo42?bar=baz HTTP/1.1\r\nHost: localhost:7878\r\nUser-Agent: Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:70.0) Gecko/20100101 Firefox/70.0\r\nAccept: image/webp,*/*\r\nAccept-Language: de-DE,en-US;q=0.7,en;q=0.3\r\nAccept-Encoding: gzip, deflate\r\n\r\n";
                             match connection.write(req) {
@@ -565,8 +576,94 @@ app = simple_app"#,
                         }
                         t.join().unwrap();
                     }
-                    Err(e) => {
-                        debug!("Error when creating Server: {:?}", e);
+                    Err(_) => {
+                        assert!(false);
+                    }
+                }
+            }
+            Err(e) => {
+                e.print_and_set_sys_last_vars(py);
+                assert!(false);
+            }
+        }
+    }
+
+    #[test]
+    fn test_server_poll_once_expect_continue() {
+        let gil = Python::acquire_gil();
+        let py = gil.python();
+        let locals = PyDict::new(py);
+        let app = py.run(
+            r#"
+def simple_app(environ, start_response):
+    status = '200 OK'
+    response_headers = [('Content-type', 'text/plain'), ]
+    start_response(status, response_headers)
+    return [b"Hello world!\n"]
+
+app = simple_app"#,
+            None,
+            Some(&locals),
+        );
+        match app {
+            Ok(_) => {
+                let (g, _) = make_globals(py);
+                let app = locals
+                    .get_item(py, "app")
+                    .unwrap()
+                    .as_object()
+                    .clone_ref(py);
+                let si: SocketAddr = "127.0.0.1:0".parse().unwrap();
+                let listener = TcpListener::bind(si).unwrap();
+                let addr = listener.local_addr().unwrap();
+                match Server::new(app, listener, server_options(g), py) {
+                    Ok(mut got) => {
+                        let (tx, rx) = channel();
+                        let (snd, rcv) = channel();
+                        // create request in separate thread
+                        let t = thread::spawn(move || {
+                            let mut connection =
+                                TcpStream::connect_timeout(&addr, Duration::from_millis(1000))
+                                    .expect("Failed to connect to server");
+                            let headers = b"POST /test HTTP/1.1\r\nHost: foo.example\r\nContent-Type: application/x-www-form-urlencoded\r\nContent-Length: 27\r\nExpect: 100-continue\r\n\r\n";
+                            let body = b"field1=value1&field2=value2";
+                            match connection.write(headers) {
+                                Ok(num_bytes) => assert_eq!(num_bytes, headers.len()),
+                                Err(_) => assert!(false),
+                            };
+                            let mut buf = [0; 25];
+                            // Attempting to read 100 Continue
+                            connection.read(&mut buf).unwrap();
+                            snd.clone().send(buf).unwrap();
+                            match connection.write(body) {
+                                Ok(num_bytes) => assert_eq!(num_bytes, body.len()),
+                                Err(_) => assert!(false),
+                            };
+                            rx.recv().unwrap();
+                        });
+                        // Listener accept
+                        if let Err(_) = got.poll_once() {
+                            assert!(false);
+                        }
+                        // Expect -> continue
+                        if let Err(_) = got.poll_once() {
+                            assert!(false);
+                        }
+                        // Read 100 continue from channel
+                        let b = rcv.recv().unwrap();
+                        assert_eq!(&b[..], b"HTTP/1.1 100 Continue\r\n\r\n");
+                        // read + propagate HTTPrequest
+                        if let Err(_) = got.poll_once() {
+                            assert!(false);
+                        }
+                        if let Err(_) = got.poll_once() {
+                            // handle request in worker
+                            assert!(false);
+                        }
+                        tx.send(()).unwrap();
+                        t.join().unwrap();
+                    }
+                    Err(_) => {
                         assert!(false);
                     }
                 }
